// TODO: is there a way to overload operators without taking ownership?
// TODO: test functions that should panic

//! # CK Libs
//! 
//! Simple libraries built by Jeremiah "CaptainKraft" Goerdt and
//! they have been built up slowly as he learns to use Rust.

use std::ops::{Add, Sub, Index, IndexMut};

#[derive(Debug)]
pub struct Vec2 {
    pub x: f32,
    pub y: f32,
}

impl Index<usize> for Vec2 {
    type Output = f32;

    /// Indexing for Vec2 instead of field access by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2 { x: 1.0, y: 2.0 };
    /// assert_eq!(v[0], 1.0);
    /// assert_eq!(v[1], 2.0);
    /// ```
    fn index(&self, idx: usize) -> &f32 {
        match idx {
            0 => &self.x,
            1 => &self.y,
            _ => panic!("Vec2 index out of bounds"),
        }
    }
}

impl IndexMut<usize> for Vec2 {
    /// Mutable indexing for Vec2 instead of setting fields by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let mut v = Vec2 { x: 0.0, y: 0.0 };
    /// v[0] = 1.0;
    /// v[1] = 2.0;
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// ```
    fn index_mut(&mut self, idx: usize) -> &mut f32 {
        match idx {
            0 => &mut self.x,
            1 => &mut self.y,
            _ => panic!("Vec2 index out of bounds"),
        }
    }
}

impl Add<Vec2> for Vec2 {
    type Output = Vec2;

    /// Overload the + operator for Vec2
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v1 = Vec2 { x: 1.0, y: 2.0 };
    /// let v2 = Vec2 { x: 3.0, y: 4.0 };
    /// let v = v1 + v2;
    ///
    /// assert_eq!(v.x, 4.0);
    /// assert_eq!(v.y, 6.0);
    /// ```
    fn add(self, other: Vec2) -> Vec2 {
        Vec2 {
            x: self.x + other.x,
            y: self.y + other.y,
        }
    }
}

impl Sub<Vec2> for Vec2 {
    type Output = Vec2;

    /// Overload the - operator for Vec2
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v1 = Vec2 { x: 1.0, y: 2.0 };
    /// let v2 = Vec2 { x: 3.0, y: 5.0 };
    /// let v = v1 - v2;
    ///
    /// assert_eq!(v.x, -2.0);
    /// assert_eq!(v.y, -3.0);
    /// ```
    fn sub(self, other: Vec2) -> Vec2 {
        Vec2 {
            x: self.x - other.x,
            y: self.y - other.y,
        }
    }
}

impl Vec2 {
    /// # Create a new instance of Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2::new(1.0, 2.0);
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// ```
    pub fn new(x: f32, y: f32) -> Vec2 {
        Vec2 { x: x, y: y }
    }

    /// # Get the length of a Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2::new(3.0, 4.0);
    /// assert_eq!(v.len(), 5.0);
    /// ```
    pub fn len(&self) -> f32 {
        let l = self.x * self.x + self.y * self.y;
        l.sqrt()
    }

    /// # Get the squared length of a Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2::new(3.0, 4.0);
    /// assert_eq!(v.len_squared(), 25.0);
    /// ```
    pub fn len_squared(&self) -> f32 {
        self.x * self.x + self.y * self.y
    }

    /// # Get a new normalized version of a Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2::new(3.0, 4.0);
    /// let vn = v.normalize();
    /// assert_eq!(vn.x, 3.0/5.0);
    /// assert_eq!(vn.y, 4.0/5.0);
    /// ```
    pub fn normalize(&self) -> Vec2 {
        let len = self.len();
        Vec2::new(self.x / len, self.y / len)
    }

    /// # Get the raw data represented by a Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v = Vec2::new(1.0, 2.0);
    /// let r: [f32; 2] = v.raw();
    /// assert_eq!(v.x, r[0]);
    /// assert_eq!(v.y, r[1]);
    /// ```
    pub fn raw(&self) -> [f32; 2] {
        [self.x, self.y]
    }

    /// # Get the dot product of two Vec2
    ///
    /// ```
    /// use ck::Vec2;
    ///
    /// let v1 = Vec2::new(1.0, 2.0);
    /// let v2 = Vec2::new(3.0, 4.0);
    /// let res: f32 = v1.dot(&v2);
    /// assert_eq!(res, 11.0);
    /// ```
    pub fn dot(&self, other: &Vec2) -> f32 {
        self.x * other.x + self.y * other.y
    }
}

#[derive(Debug)]
pub struct Vec3 {
    pub x: f32,
    pub y: f32,
    pub z: f32,
}

impl Index<usize> for Vec3 {
    type Output = f32;

    /// Indexing for Vec3 instead of field access by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3 { x: 1.0, y: 2.0, z: 3.0 };
    /// assert_eq!(v[0], 1.0);
    /// assert_eq!(v[1], 2.0);
    /// assert_eq!(v[2], 3.0);
    /// ```
    fn index(&self, idx: usize) -> &f32 {
        match idx {
            0 => &self.x,
            1 => &self.y,
            2 => &self.z,
            _ => panic!("Vec3 index out of bounds"),
        }
    }
}

impl IndexMut<usize> for Vec3 {
    /// Mutable indexing for Vec3 instead of setting fields by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let mut v = Vec3 { x: 0.0, y: 0.0, z: 0.0 };
    /// v[0] = 1.0;
    /// v[1] = 2.0;
    /// v[2] = 3.0;
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// assert_eq!(v.z, 3.0);
    /// ```
    fn index_mut(&mut self, idx: usize) -> &mut f32 {
        match idx {
            0 => &mut self.x,
            1 => &mut self.y,
            2 => &mut self.z,
            _ => panic!("Vec3 index out of bounds"),
        }
    }
}

impl Add<Vec3> for Vec3 {
    type Output = Vec3;

    /// Overload the + operator for Vec3
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v1 = Vec3 { x: 1.0, y: 2.0, z: 3.0 };
    /// let v2 = Vec3 { x: 4.0, y: 5.0, z: 6.0 };
    /// let v = v1 + v2;
    ///
    /// assert_eq!(v.x, 5.0);
    /// assert_eq!(v.y, 7.0);
    /// assert_eq!(v.z, 9.0);
    /// ```
    fn add(self, other: Vec3) -> Vec3 {
        Vec3 {
            x: self.x + other.x,
            y: self.y + other.y,
            z: self.z + other.z,
        }
    }
}

impl Sub<Vec3> for Vec3 {
    type Output = Vec3;

    /// Overload the - operator for Vec3
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v1 = Vec3 { x: 1.0, y: 2.0, z: 3.0 };
    /// let v2 = Vec3 { x: 6.0, y: 5.0, z: 4.0 };
    /// let v = v1 - v2;
    ///
    /// assert_eq!(v.x, -5.0);
    /// assert_eq!(v.y, -3.0);
    /// assert_eq!(v.z, -1.0);
    /// ```
    fn sub(self, other: Vec3) -> Vec3 {
        Vec3 {
            x: self.x - other.x,
            y: self.y - other.y,
            z: self.z - other.z,
        }
    }
}

impl Vec3 {
    /// # Create a new instance of Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3::new(1.0, 2.0, 3.0);
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// assert_eq!(v.z, 3.0);
    /// ```
    pub fn new(x: f32, y: f32, z: f32) -> Vec3 {
        Vec3 { x: x, y: y, z: z }
    }

    /// # Get the length of a Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3::new(2.0, 4.0, 4.0);
    /// assert_eq!(v.len(), 6.0);
    /// ```
    pub fn len(&self) -> f32 {
        let l = self.x * self.x + self.y * self.y + self.z * self.z;
        l.sqrt()
    }

    /// # Get the squared length of a Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3::new(2.0, 4.0, 4.0);
    /// assert_eq!(v.len_squared(), 36.0);
    /// ```
    pub fn len_squared(&self) -> f32 {
        self.x * self.x + self.y * self.y + self.z * self.z
    }

    /// # Get a new normalized version of a Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3::new(2.0, 4.0, 4.0);
    /// let vn = v.normalize();
    /// assert_eq!(vn.x, 2.0/6.0);
    /// assert_eq!(vn.y, 4.0/6.0);
    /// assert_eq!(vn.z, 4.0/6.0);
    /// ```
    pub fn normalize(&self) -> Vec3 {
        let len = self.len();
        Vec3::new(self.x / len, self.y / len, self.z / len)
    }

    /// # Get the raw data represented by a Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v = Vec3::new(1.0, 2.0, 3.0);
    /// let r: [f32; 3] = v.raw();
    /// assert_eq!(v.x, r[0]);
    /// assert_eq!(v.y, r[1]);
    /// assert_eq!(v.z, r[2]);
    /// ```
    pub fn raw(&self) -> [f32; 3] {
        [self.x, self.y, self.z]
    }

    /// # Get the dot product of two Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v1 = Vec3::new(1.0, 2.0, 3.0);
    /// let v2 = Vec3::new(4.0, 5.0, 6.0);
    /// let res: f32 = v1.dot(&v2);
    /// assert_eq!(res, 32.0);
    /// ```
    pub fn dot(&self, other: &Vec3) -> f32 {
        self.x * other.x + self.y * other.y + self.z * other.z
    }

    /// # Get the cross product of two Vec3
    ///
    /// ```
    /// use ck::Vec3;
    ///
    /// let v1 = Vec3::new(1.0, 2.0, 3.0);
    /// let v2 = Vec3::new(4.0, 5.0, 6.0);
    /// let v = v1.cross(&v2);
    /// assert_eq!(v.x, -3.0);
    /// assert_eq!(v.y,  6.0);
    /// assert_eq!(v.z, -3.0);
    /// ```
    pub fn cross(&self, other: &Vec3) -> Vec3 {
        Vec3 {
            x: self.y * other.z - self.z * other.y,
            y: self.z * other.x - self.x * other.z,
            z: self.x * other.y - self.y * other.x,
        }
    }
}

#[derive(Debug)]
pub struct Vec4 {
    pub x: f32,
    pub y: f32,
    pub z: f32,
    pub w: f32,
}

impl Index<usize> for Vec4 {
    type Output = f32;

    /// Indexing for Vec4 instead of field access by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4 { x: 1.0, y: 2.0, z: 3.0, w: 4.0 };
    /// assert_eq!(v[0], 1.0);
    /// assert_eq!(v[1], 2.0);
    /// assert_eq!(v[2], 3.0);
    /// assert_eq!(v[3], 4.0);
    /// ```
    fn index(&self, idx: usize) -> &f32 {
        match idx {
            0 => &self.x,
            1 => &self.y,
            2 => &self.z,
            3 => &self.w,
            _ => panic!("Vec4 index out of bounds"),
        }
    }
}

impl IndexMut<usize> for Vec4 {
    /// Mutable indexing for Vec4 instead of setting fields by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let mut v = Vec4 { x: 0.0, y: 0.0, z: 0.0, w: 0.0 };
    /// v[0] = 1.0;
    /// v[1] = 2.0;
    /// v[2] = 3.0;
    /// v[3] = 4.0;
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// assert_eq!(v.z, 3.0);
    /// assert_eq!(v.w, 4.0);
    /// ```
    fn index_mut(&mut self, idx: usize) -> &mut f32 {
        match idx {
            0 => &mut self.x,
            1 => &mut self.y,
            2 => &mut self.z,
            3 => &mut self.w,
            _ => panic!("Vec4 index out of bounds"),
        }
    }
}

impl Add<Vec4> for Vec4 {
    type Output = Vec4;

    /// Overload the + operator for Vec4
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v1 = Vec4 { x: 1.0, y: 2.0, z: 3.0, w: 4.0 };
    /// let v2 = Vec4 { x: 5.0, y: 6.0, z: 7.0, w: 8.0 };
    /// let v = v1 + v2;
    ///
    /// assert_eq!(v.x,  6.0);
    /// assert_eq!(v.y,  8.0);
    /// assert_eq!(v.z, 10.0);
    /// assert_eq!(v.w, 12.0);
    /// ```
    fn add(self, other: Vec4) -> Vec4 {
        Vec4 {
            x: self.x + other.x,
            y: self.y + other.y,
            z: self.z + other.z,
            w: self.w + other.w,
        }
    }
}

impl Sub<Vec4> for Vec4 {
    type Output = Vec4;

    /// Overload the - operator for Vec4
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v1 = Vec4 { x: 1.0, y: 2.0, z: 3.0, w: 4.0 };
    /// let v2 = Vec4 { x: 8.0, y: 7.0, z: 6.0, w: 5.0 };
    /// let v = v1 - v2;
    ///
    /// assert_eq!(v.x, -7.0);
    /// assert_eq!(v.y, -5.0);
    /// assert_eq!(v.z, -3.0);
    /// assert_eq!(v.w, -1.0);
    /// ```
    fn sub(self, other: Vec4) -> Vec4 {
        Vec4 {
            x: self.x - other.x,
            y: self.y - other.y,
            z: self.z - other.z,
            w: self.w - other.w,
        }
    }
}

impl Vec4 {
    /// # Create a new instance of Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4::new(1.0, 2.0, 3.0, 4.0);
    /// assert_eq!(v.x, 1.0);
    /// assert_eq!(v.y, 2.0);
    /// assert_eq!(v.z, 3.0);
    /// assert_eq!(v.w, 4.0);
    /// ```
    pub fn new(x: f32, y: f32, z: f32, w: f32) -> Vec4 {
        Vec4 { x: x, y: y, z: z, w: w }
    }

    /// # Get the length of a Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4::new(1.0, 2.0, 3.0, 5.0);
    /// let a = v.len() * 1000.0;
    ///
    /// // Rounding the result to account for f32 precision.
    /// assert_eq!(a.round(), 6245.0);
    /// ```
    pub fn len(&self) -> f32 {
        let l = self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w;
        l.sqrt()
    }

    /// # Get the squared length of a Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4::new(1.0, 2.0, 3.0, 5.0);
    /// assert_eq!(v.len_squared(), 39.0);
    /// ```
    pub fn len_squared(&self) -> f32 {
        self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w
    }

    /// # Get a new normalized version of a Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4::new(1.0, 2.0, 3.0, 5.0);
    /// let vn = v.normalize();
    ///
    /// assert_eq!(vn.x, 1.0/v.len());
    /// assert_eq!(vn.y, 2.0/v.len());
    /// assert_eq!(vn.z, 3.0/v.len());
    /// assert_eq!(vn.w, 5.0/v.len());
    /// ```
    pub fn normalize(&self) -> Vec4 {
        let len = self.len();
        Vec4::new(self.x / len, self.y / len, self.z / len, self.w / len)
    }

    /// # Get the raw data represented by a Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v = Vec4::new(1.0, 2.0, 3.0, 4.0);
    /// let r: [f32; 4] = v.raw();
    /// assert_eq!(v.x, r[0]);
    /// assert_eq!(v.y, r[1]);
    /// assert_eq!(v.z, r[2]);
    /// assert_eq!(v.w, r[3]);
    /// ```
    pub fn raw(&self) -> [f32; 4] {
        [self.x, self.y, self.z, self.w]
    }

    /// # Get the dot product of two Vec4
    ///
    /// ```
    /// use ck::Vec4;
    ///
    /// let v1 = Vec4::new(1.0, 2.0, 3.0, 4.0);
    /// let v2 = Vec4::new(5.0, 6.0, 7.0, 8.0);
    /// let res: f32 = v1.dot(&v2);
    /// assert_eq!(res, 70.0);
    /// ```
    pub fn dot(&self, other: &Vec4) -> f32 {
        self.x * other.x + self.y * other.y + self.z * other.z + self.w * other.w
    }
}

#[derive(Debug)]
pub struct Mat4 {
    pub elems: [Vec4; 4],
}

impl Index<usize> for Mat4 {
    type Output = Vec4;

    /// Indexing for Mat4 instead of field access by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let v1 = Vec4::new( 1.0,  2.0,  3.0,  4.0);
    /// let v2 = Vec4::new( 5.0,  6.0,  7.0,  8.0);
    /// let v3 = Vec4::new( 9.0, 10.0, 11.0, 12.0);
    /// let v4 = Vec4::new(13.0, 14.0, 15.0, 16.0);
    /// let m = Mat4 { elems: [v1, v2, v3, v4] };
    ///
    /// assert_eq!(m[0].x, 1.0);
    /// assert_eq!(m[0].y, 2.0);
    /// assert_eq!(m[0].z, 3.0);
    /// assert_eq!(m[0].w, 4.0);
    ///
    /// assert_eq!(m[1][0], 5.0);
    /// assert_eq!(m[1][1], 6.0);
    /// assert_eq!(m[1][2], 7.0);
    /// assert_eq!(m[1][3], 8.0);
    ///
    /// assert_eq!(m[2].x,  9.0);
    /// assert_eq!(m[2].y, 10.0);
    /// assert_eq!(m[2].z, 11.0);
    /// assert_eq!(m[2].w, 12.0);
    ///
    /// assert_eq!(m[3][0], 13.0);
    /// assert_eq!(m[3][1], 14.0);
    /// assert_eq!(m[3][2], 15.0);
    /// assert_eq!(m[3][3], 16.0);
    /// ```
    fn index(&self, idx: usize) -> &Vec4 {
        match idx {
            0 ... 3 => &self.elems[idx],
            _ => panic!("Mat4 index out of bounds"),
        }
    }
}

impl IndexMut<usize> for Mat4 {
    /// Mutable indexing for Mat4 instead of field access by name
    ///
    /// # Examples
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let mut m = Mat4::new(
    ///     Vec4::new(0.0, 0.0, 0.0, 0.0),
    ///     Vec4::new(0.0, 0.0, 0.0, 0.0),
    ///     Vec4::new(0.0, 0.0, 0.0, 0.0),
    ///     Vec4::new(0.0, 0.0, 0.0, 0.0),
    /// );
    /// let v1 = Vec4::new( 1.0,  2.0,  3.0,  4.0);
    /// let v2 = Vec4::new( 5.0,  6.0,  7.0,  8.0);
    /// let v3 = Vec4::new( 9.0, 10.0, 11.0, 12.0);
    /// let v4 = Vec4::new(13.0, 14.0, 15.0, 16.0);
    ///
    /// m[0] = v1;
    /// m[1] = v2;
    /// m[2] = v3;
    /// m[3] = v4;
    ///
    /// assert_eq!(m[0].x, 1.0);
    /// assert_eq!(m[0].y, 2.0);
    /// assert_eq!(m[0].z, 3.0);
    /// assert_eq!(m[0].w, 4.0);
    ///
    /// assert_eq!(m[1][0], 5.0);
    /// assert_eq!(m[1][1], 6.0);
    /// assert_eq!(m[1][2], 7.0);
    /// assert_eq!(m[1][3], 8.0);
    ///
    /// assert_eq!(m[2].x,  9.0);
    /// assert_eq!(m[2].y, 10.0);
    /// assert_eq!(m[2].z, 11.0);
    /// assert_eq!(m[2].w, 12.0);
    ///
    /// assert_eq!(m[3][0], 13.0);
    /// assert_eq!(m[3][1], 14.0);
    /// assert_eq!(m[3][2], 15.0);
    /// assert_eq!(m[3][3], 16.0);
    /// ```
    fn index_mut(&mut self, idx: usize) -> &mut Vec4 {
        match idx {
            0 ... 3 => &mut self.elems[idx],
            _ => panic!("Mat4 index out of bounds"),
        }
    }
}

impl Mat4 {
    /// # Create a new instance of Mat4
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let v1 = Vec4::new( 1.0,  2.0,  3.0,  4.0);
    /// let v2 = Vec4::new( 5.0,  6.0,  7.0,  8.0);
    /// let v3 = Vec4::new( 9.0, 10.0, 11.0, 12.0);
    /// let v4 = Vec4::new(13.0, 14.0, 15.0, 16.0);
    /// let m = Mat4::new(v1, v2, v3, v4);
    ///
    /// assert_eq!(m[0].x, 1.0);
    /// assert_eq!(m[0].y, 2.0);
    /// assert_eq!(m[0].z, 3.0);
    /// assert_eq!(m[0].w, 4.0);
    ///
    /// assert_eq!(m[1][0], 5.0);
    /// assert_eq!(m[1][1], 6.0);
    /// assert_eq!(m[1][2], 7.0);
    /// assert_eq!(m[1][3], 8.0);
    ///
    /// assert_eq!(m[2].x,  9.0);
    /// assert_eq!(m[2].y, 10.0);
    /// assert_eq!(m[2].z, 11.0);
    /// assert_eq!(m[2].w, 12.0);
    ///
    /// assert_eq!(m[3][0], 13.0);
    /// assert_eq!(m[3][1], 14.0);
    /// assert_eq!(m[3][2], 15.0);
    /// assert_eq!(m[3][3], 16.0);
    /// ```
    pub fn new(a: Vec4, b: Vec4, c: Vec4, d: Vec4) -> Mat4 {
        Mat4 { elems: [a, b, c, d] }
    }

    /// # Create a new instance of the identity Mat4
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let m = Mat4::identity();
    ///
    /// assert_eq!(m[0][0], 1.0);
    /// assert_eq!(m[0][1], 0.0);
    /// assert_eq!(m[0][2], 0.0);
    /// assert_eq!(m[0][3], 0.0);
    ///
    /// assert_eq!(m[1][0], 0.0);
    /// assert_eq!(m[1][1], 1.0);
    /// assert_eq!(m[1][2], 0.0);
    /// assert_eq!(m[1][3], 0.0);
    ///
    /// assert_eq!(m[2][0], 0.0);
    /// assert_eq!(m[2][1], 0.0);
    /// assert_eq!(m[2][2], 1.0);
    /// assert_eq!(m[2][3], 0.0);
    ///
    /// assert_eq!(m[3][0], 0.0);
    /// assert_eq!(m[3][1], 0.0);
    /// assert_eq!(m[3][2], 0.0);
    /// assert_eq!(m[3][3], 1.0);
    /// ```
    pub fn identity() -> Mat4 {
        let a = Vec4::new(1.0, 0.0, 0.0, 0.0);
        let b = Vec4::new(0.0, 1.0, 0.0, 0.0);
        let c = Vec4::new(0.0, 0.0, 1.0, 0.0);
        let d = Vec4::new(0.0, 0.0, 0.0, 1.0);

        Mat4 { elems: [a, b, c, d] }
    }

    /// # Create a new instance of the zero Mat4
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let m = Mat4::zero();
    ///
    /// assert_eq!(m[0][0], 0.0);
    /// assert_eq!(m[0][1], 0.0);
    /// assert_eq!(m[0][2], 0.0);
    /// assert_eq!(m[0][3], 0.0);
    ///
    /// assert_eq!(m[1][0], 0.0);
    /// assert_eq!(m[1][1], 0.0);
    /// assert_eq!(m[1][2], 0.0);
    /// assert_eq!(m[1][3], 0.0);
    ///
    /// assert_eq!(m[2][0], 0.0);
    /// assert_eq!(m[2][1], 0.0);
    /// assert_eq!(m[2][2], 0.0);
    /// assert_eq!(m[2][3], 0.0);
    ///
    /// assert_eq!(m[3][0], 0.0);
    /// assert_eq!(m[3][1], 0.0);
    /// assert_eq!(m[3][2], 0.0);
    /// assert_eq!(m[3][3], 0.0);
    /// ```
    pub fn zero() -> Mat4 {
        let a = Vec4::new(0.0, 0.0, 0.0, 0.0);
        let b = Vec4::new(0.0, 0.0, 0.0, 0.0);
        let c = Vec4::new(0.0, 0.0, 0.0, 0.0);
        let d = Vec4::new(0.0, 0.0, 0.0, 0.0);

        Mat4 { elems: [a, b, c, d] }
    }

    pub fn ortho(l: f32, r: f32, b: f32, t: f32, n: f32, f: f32) -> Mat4 {
        let mut m = Mat4::identity();

        m[0][0] = 2.0 / (r - l);
        m[1][1] = 2.0 / (t - b);
        m[2][2] = 2.0 / (n - f);

        m[3][0] = (l + r) / (l - r);
        m[3][1] = (b + t) / (b - t);
        m[3][2] = (f + n) / (n - f);

        m
    }

    /// # Get the raw data represented by a Mat4
    ///
    /// ```
    /// use ck::{Vec4, Mat4};
    ///
    /// let v1 = Vec4::new( 1.0,  2.0,  3.0,  4.0);
    /// let v2 = Vec4::new( 5.0,  6.0,  7.0,  8.0);
    /// let v3 = Vec4::new( 9.0, 10.0, 11.0, 12.0);
    /// let v4 = Vec4::new(13.0, 14.0, 15.0, 16.0);
    /// let m = Mat4::new(v1, v2, v3, v4);
    /// let r: [[f32; 4]; 4] = m.raw();
    ///
    /// assert_eq!(m[0][0], r[0][0]);
    /// assert_eq!(m[0][1], r[0][1]);
    /// assert_eq!(m[0][2], r[0][2]);
    /// assert_eq!(m[0][3], r[0][3]);
    ///
    /// assert_eq!(m[1][0], r[1][0]);
    /// assert_eq!(m[1][1], r[1][1]);
    /// assert_eq!(m[1][2], r[1][2]);
    /// assert_eq!(m[1][3], r[1][3]);
    ///
    /// assert_eq!(m[2][0], r[2][0]);
    /// assert_eq!(m[2][1], r[2][1]);
    /// assert_eq!(m[2][2], r[2][2]);
    /// assert_eq!(m[2][3], r[2][3]);
    ///
    /// assert_eq!(m[3][0], r[3][0]);
    /// assert_eq!(m[3][1], r[3][1]);
    /// assert_eq!(m[3][2], r[3][2]);
    /// assert_eq!(m[3][3], r[3][3]);
    /// ```
    pub fn raw(&self) -> [[f32; 4]; 4] {
        [self[0].raw(), self[1].raw(), self[2].raw(), self[3].raw()]
    }
}

#[cfg(test)]
mod tests {
    #[test]
    fn it_works() {
        assert_eq!(2 + 2, 4);
    }
}
